@@ -1,3 +1,4 @@
+/* identify.c - by Mark Lord (C) 2000-2007 -- freely distributable */
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
@@ -99,6 +100,7 @@
 #define LENGTH_MEDIA            20  /* 20 words (40 bytes or characters)*/
 #define START_MANUF             196 /* media manufacturer I.D. */
 #define LENGTH_MANUF            10  /* 10 words (20 bytes or characters) */
+#define SCT_SUPP		206 /* SMART command transport (SCT) support */
 #define TRANSPORT_MAJOR		222 /* PATA vs. SATA etc.. */
 #define TRANSPORT_MINOR		223 /* minor revision number */
 #define INTEGRITY		255 /* integrity word */
@@ -272,11 +274,11 @@ const char *minor_str[MINOR_MAX+2] = {			/* word 81 value: */
 	"ATA/ATAPI-6 T13 1410D revision 1",		/* 0x001c	*/
 	"ATA/ATAPI-7 published, ANSI INCITS 397-2005",	/* 0x001d	*/
 	"ATA/ATAPI-7 T13 1532D revision 0",		/* 0x001e	*/
-	"Reserved"					/* 0x001f	*/
-	"Reserved"					/* 0x0020	*/
+	"Reserved",					/* 0x001f	*/
+	"Reserved",					/* 0x0020	*/
 	"ATA/ATAPI-7 T13 1532D revision 4a",		/* 0x0021	*/
 	"ATA/ATAPI-6 published, ANSI INCITS 361-2002",	/* 0x0022	*/
-	"Reserved"					/* 0x0023-0xfffe*/
+	"Reserved",					/* 0x0023-0xfffe*/
 };
 const char actual_ver[MINOR_MAX+2] = { 
 			/* word 81 value: */
@@ -329,7 +331,7 @@ static const char unknown[8] = "obsolete";
 //static const char unknown[8] = "unknown";
 #define unknown "unknown-"
 
-static const char *feat_0_str[16] = { 
+static const char *feat_word82_str[16] = { 
 	"obsolete 82[15]",				/* word 82 bit 15: obsolete  */
 	"NOP cmd",					/* word 82 bit 14 */
 	"READ_BUFFER command",				/* word 82 bit 13 */
@@ -347,7 +349,7 @@ static const char *feat_0_str[16] = {
 	"Security Mode feature set",			/* word 82 bit  1 */
 	"SMART feature set"				/* word 82 bit  0 */
 };
-static const char *feat_1_str[16] = { 
+static const char *feat_word83_str[16] = { 
 	NULL,						/* word 83 bit 15: !valid bit */
 	NULL,						/* word 83 bit 14:  valid bit */
 	"FLUSH_CACHE_EXT",				/* word 83 bit 13 */
@@ -365,12 +367,12 @@ static const char *feat_1_str[16] = {
 	"READ/WRITE_DMA_QUEUED",			/* word 83 bit  1 */
 	"DOWNLOAD_MICROCODE"				/* word 83 bit  0 */
 };
-static const char *feat_2_str[16] = { 
+static const char *feat_word84_str[16] = { 
 	NULL,						/* word 84 bit 15: !valid bit */
 	NULL,						/* word 84 bit 14:  valid bit */
 	"IDLE_IMMEDIATE with UNLOAD",			/* word 84 bit 13 */
-	"unknown 84[12]",				/* word 84 bit 12 */
-	"unknown 84[11]",				/* word 84 bit 11 */
+	"Command Completion Time Limit (CCTL)",		/* word 84 bit 12 (ref: dt1696) */
+	"Time Limited Commands (TLC) feature set",	/* word 84 bit 11 (ref: dt1696) */
 	"URG for WRITE_STREAM[_DMA]_EXT",		/* word 84 bit 10 */
 	"URG for READ_STREAM[_DMA]_EXT",		/* word 84 bit  9 */
 	"64-bit World wide name",			/* word 84 bit  8 */
@@ -399,7 +401,7 @@ static const char *feat_3_str[16] = {
 	"{READ,WRITE}_DMA_EXT_GPL commands",		/* word 119 bit  3 */
 	"WRITE_UNCORRECTABLE command",			/* word 119 bit  2 */
 	"Write-Read-Verify feature set",		/* word 119 bit  1 */
-	"unknown 119[0]"				/* word 119 bit  0: reserved for DT2014 */
+	"Disable Data Transfer After Error Detection"	/* word 119 bit  0 (ref: 2014DT)*/
 };
 static const char *cap_sata0_str[16] = { 
 	"unknown 76[15]",				/* word 76 bit 15 */
@@ -430,7 +432,7 @@ static const char *feat_sata0_str[16] = {
 	"unknown 78[8]",				/* word 78 bit  8 */
 	"unknown 78[7]",				/* word 78 bit  7 */
 	"Software settings preservation",		/* word 78 bit  6 */
-	"unknown 78[5]",				/* word 78 bit  5 */
+	"Asynchronous notification (eg. media change)",	/* word 78 bit  5 */
 	"In-order data delivery",			/* word 78 bit  4 */
 	"Device-initiated interface power management",	/* word 78 bit  3 */
 	"DMA Setup Auto-Activate optimization",		/* word 78 bit  2 */
@@ -479,6 +481,26 @@ const char *secu_str[] = {
 #define PWR_MODE_OFF		0x1000  /* 1=CFA power moded disabled */
 #define MAX_AMPS		0x0fff  /* value = max current in ma */
 
+/* word 206: SMART command transport (SCT) */
+static const char *feat_sct_str[16] = {
+	"unknown 206[15]",				/* word 206 bit 15 */
+	"unknown 206[14]",				/* word 206 bit 14 */
+	"unknown 206[13]",				/* word 206 bit 13 */
+	"unknown 206[12]",				/* word 206 bit 12 */
+	"unknown 206[11]",				/* word 206 bit 11 */
+	"unknown 206[10]",				/* word 206 bit 10 */
+	"unknown 206[9]",				/* word 206 bit  9 */
+	"unknown 206[8]",				/* word 206 bit  8 */
+	"unknown 206[7]",				/* word 206 bit  7 */
+	"unknown 206[6]",				/* word 206 bit  6 */
+	"SCT Data Tables (AC5)",			/* word 206 bit  5 */
+	"SCT Features Control (AC4)",			/* word 206 bit  4 */
+	"SCT Error Recovery Control (AC3)",		/* word 206 bit  3 */
+	"SCT LBA Segment Access (AC2)",			/* word 206 bit  2 */
+	"SCT Long Sector Access (AC1)",			/* word 206 bit  1 */
+	"SMART Command Transport (SCT) feature set"	/* word 206 bit  0 */
+};
+
 /* word 255: integrity */
 #define SIG			0x00ff  /* signature location */
 #define SIG_VAL			0x00A5  /* signature value */
@@ -511,13 +533,23 @@ static void print_features (__u16 supported, __u16 enabled, const char *names[])
 	}
 }
 
-static int print_transport_type(__u16 major, __u16 minor)
+static int print_transport_type(__u16 val[])
 {
+	__u16 major = val[TRANSPORT_MAJOR], minor = val[TRANSPORT_MINOR];
 	unsigned int ttype, subtype, transport = 0;
 
-	if (major == 0x0000 || major == 0xffff)
+	if (major == 0x0000 || major == 0xffff) {
+#if 0
+		printf("\t%-20snot reported","Transport:");
+		if ((val[SATA_CAP_0]  && val[SATA_CAP_0]  != 0xffff)
+		 || (val[SATA_SUPP_0] && val[SATA_SUPP_0] != 0xffff)) {
+			printf(" (serial)");
+		}
+		putchar('\n');
+#endif
 		return transport;
-	printf("Transport: ");
+	}
+	printf("\t%-20s","Transport:");
 	ttype = major >> 12;
 	subtype = major & 0xfff;
 	transport = ttype;
@@ -559,11 +591,9 @@ static int print_transport_type(__u16 major, __u16 minor)
 }
 
 /* our main() routine: */
-void identify (__u16 *id_supplied, const char *devname)
+void identify (__u16 *id_supplied)
 {
 
-	char *id_file = NULL, fmt[]="/proc/ide/%s/identify";
-	FILE *fl;
 	__u16 val[256], ii, jj, kk;
 	__u16 like_std = 1, std = 0, min_std = 0xffff;
 	__u16 dev = NO_DEV, eqpt = NO_DEV;
@@ -573,33 +603,12 @@ void identify (__u16 *id_supplied, const char *devname)
 	__u64 bb, bbbig; /* (:) */
 	int transport;
 
-	if (id_supplied) {
-		memcpy(val, id_supplied, sizeof(val));
-	} else {
-		id_file = calloc(1, strlen(devname)+1+strlen(fmt));
-		sprintf(id_file, fmt, devname);
+	memcpy(val, id_supplied, sizeof(val));
 
-		/* open the file, read in all the info and close it */
-		if (id_file == NULL) {
-			fl = stdin;
-		} else if(NULL == (fl = fopen(id_file, "r"))) {
-     			perror(id_file);
-			exit(errno);
-		}
-		/* calculate checksum over all bytes */
-		for(ii = GEN_CONFIG; ii<=INTEGRITY; ii++) {
-			unsigned int scratch;
-			if(1 != fscanf(fl,"%04x",&scratch)) break;
-			val[ii] = (__u16)scratch;
-			chksum += val[ii] + (val[ii] >> 8);
-		}
-		fclose(fl);  
-		if(ii < (INTEGRITY+1)) {
-			fprintf(stderr,"Input file wrong format or length\n");
-			exit(ii);
-		}
+	/* calculate checksum over all bytes */
+	for (ii = GEN_CONFIG; ii<=INTEGRITY; ii++) {
+		chksum += val[ii] + (val[ii] >> 8);
 	}
-	chksum &= 0xff;
 
 	/* check if we recognise the device type */
 	printf("\n");
@@ -664,7 +673,7 @@ void identify (__u16 *id_supplied, const char *devname)
 		print_ascii(&val[START_MANUF], LENGTH_MANUF);
 	}
 
-	transport = print_transport_type(val[TRANSPORT_MAJOR], val[TRANSPORT_MINOR]);
+	transport = print_transport_type(val);
 
 	/* major & minor standards version number (Note: these words were not
 	 * defined until ATA-3 & the CDROM std uses different words.) */
@@ -688,7 +697,7 @@ void identify (__u16 *id_supplied, const char *devname)
 			printf("\n\tSupported: ");
 			jj = val[MAJOR] << 1;
 			kk = min_ata_std(like_std);
-			for(ii = 14; ii > kk; ii--) {
+			for (ii = 14; ii > kk; ii--) {
 				if(jj & 0x8000) {
 					printf("%u ", ii);
 					if (ii > like_std) {
@@ -752,7 +761,7 @@ void identify (__u16 *id_supplied, const char *devname)
 			kk = 1;
 			printf("\n\tSupported: CD-ROM ATAPI");
 			jj = val[CDR_MAJOR] >> 1;
-			for(ii = 1; ii <15; ii++) {
+			for (ii = 1; ii <15; ii++) {
 				if(jj & 0x0001) {
 					printf("-%u ", ii);
 				}
@@ -772,7 +781,7 @@ void identify (__u16 *id_supplied, const char *devname)
 	/* more info from the general configuration word */
 	if((eqpt != CDROM) && (like_std == 1)) {
 		jj = val[GEN_CONFIG] >> 1;
-		for(ii = 1; ii < 15; ii++) {
+		for (ii = 1; ii < 15; ii++) {
 			if(jj & 0x0001) printf("\t%s\n",ata1_cfg_str[ii]);
 			jj >>=1;
 		}
@@ -999,14 +1008,14 @@ void identify (__u16 *id_supplied, const char *devname)
 	 * than n (e.g. 3, 2, 1 and 0).  Print all the modes. */
 	if((val[WHATS_VALID] & OK_W64_70) && (val[ADV_PIO_MODES] & PIO_SUP)) {
 		jj = ((val[ADV_PIO_MODES] & PIO_SUP) << 3) | 0x0007;
-		for(ii = 0; ii <= PIO_MODE_MAX ; ii++) {
+		for (ii = 0; ii <= PIO_MODE_MAX ; ii++) {
 			if(jj & 0x0001)
 				printf("pio%d ",ii);
 			jj >>=1;
 		}
 		printf("\n");
 	} else if(((min_std < 5) || (eqpt == CDROM)) && ((val[PIO_MODE]>>8) <= 2)) {
-		for(ii = 0; ii <= val[PIO_MODE]>>8; ii++) {
+		for (ii = 0; ii <= val[PIO_MODE]>>8; ii++) {
 			printf("pio%d ",ii);
 		}
 		printf("\n");
@@ -1024,12 +1033,15 @@ void identify (__u16 *id_supplied, const char *devname)
 
 	if((val[CMDS_SUPP_1] & VALID) == VALID_VAL){
 		printf("Commands/features:\n\tEnabled\tSupported:\n");
-		print_features(val[CMDS_SUPP_0] & 0x7fff, val[CMDS_EN_0], feat_0_str);
+		print_features(val[CMDS_SUPP_0] & 0x7fff, val[CMDS_EN_0], feat_word82_str);
 		if( (val[CMDS_SUPP_1] &  VALID) == VALID_VAL)
-			print_features(val[CMDS_SUPP_1] & 0x3fff, val[CMDS_EN_1], feat_1_str);
+			print_features(val[CMDS_SUPP_1] & 0x3fff, val[CMDS_EN_1], feat_word83_str);
 		if( (val[CMDS_SUPP_2] &  VALID) == VALID_VAL
-		 && (val[CMDS_EN_2]  &   VALID) == VALID_VAL)
-			print_features(val[CMDS_SUPP_2] & 0x3fff, val[CMDS_EN_2], feat_2_str);
+		 && (val[CMDS_EN_2]  &   VALID) == VALID_VAL) {
+			print_features(val[CMDS_SUPP_2] & 0x3fff, val[CMDS_EN_2], feat_word84_str);
+			if ((val[CMDS_SUPP_2] & 0x1800) == 0x1800 && val[116] && val[116] != 0xffff)
+				printf("                (%u msec for TLC completion timer)\n", 10 * (unsigned int)(val[116]));
+		}
 		if( (val[CMDS_SUPP_1] &  VALID) == VALID_VAL
 		 && (val[CMDS_EN_1]   & 0x8000) == 0x8000
 		 && (val[CMDS_SUPP_3] &  VALID) == VALID_VAL
@@ -1039,19 +1051,21 @@ void identify (__u16 *id_supplied, const char *devname)
 			print_features(val[SATA_CAP_0],  val[SATA_CAP_0], cap_sata0_str);
 		if (transport == 1 || (val[SATA_SUPP_0] && val[SATA_SUPP_0] != 0xffff))
 			print_features(val[SATA_SUPP_0], val[SATA_EN_0], feat_sata0_str);
+		if (val[SCT_SUPP] & 0x1)
+			print_features(val[SCT_SUPP], val[SCT_SUPP] & 0x3f, feat_sct_str);
 	}
 	if((val[RM_STAT] & RM_STAT_BITS) == RM_STAT_SUP) 
 		printf("\tRemovable Media Status Notification feature set supported\n");
 
 	/* security */
-	if((eqpt != CDROM) && (like_std > 3) && 
-	   (val[SECU_STATUS] || val[ERASE_TIME] || val[ENH_ERASE_TIME])) {
+	if((eqpt != CDROM) && (like_std > 3) && (val[SECU_STATUS] || val[ERASE_TIME] || val[ENH_ERASE_TIME]))
+	{
 		printf("Security: \n");
 		if(val[PSWD_CODE] && (val[PSWD_CODE] != NOVAL_1))
 			printf("\tMaster password revision code = %u\n",val[PSWD_CODE]);
 		jj = val[SECU_STATUS];
 		if(jj) {
-			for(ii = 0; ii < NUM_SECU_STR; ii++) {
+			for (ii = 0; ii < NUM_SECU_STR; ii++) {
 				if(!(jj & 0x0001)) printf("\tnot\t");
 				else		   printf("\t\t");
 				printf("%s\n",secu_str[ii]);
@@ -1106,9 +1120,13 @@ void identify (__u16 *id_supplied, const char *devname)
 				printf("\tMaximum current = %uma\n",val[CFA_PWR_MODE] & MAX_AMPS);
 		}
 		if((val[INTEGRITY] & SIG) == SIG_VAL) {
-			printf("Checksum: ");
-			if(chksum) printf("in");
-			printf("correct\n");
+			printf("Checksum: %scorrect", chksum ? "in" : "");
+			if (chksum)
+				printf(" (0x%02x), expected 0x%02x\n", chksum, 0x100 - chksum);
+			putchar('\n');
+		} else {
+			printf("\tIntegrity word not set (found 0x%04x, expected 0x%02x%02x)\n",
+				val[INTEGRITY], 0x100 - chksum, SIG_VAL);
 		}
 	}
 }
@@ -1116,7 +1134,7 @@ void identify (__u16 *id_supplied, const char *devname)
 __u8 mode_loop(__u16 mode_sup, __u16 mode_sel, int cc, __u8 *have_mode) {
 	__u16 ii;
 	__u8 err_dma = 0;
-	for(ii = 0; ii <= MODE_MAX; ii++) {
+	for (ii = 0; ii <= MODE_MAX; ii++) {
 		if(mode_sel & 0x0001) {
 			printf("*%cdma%u ",cc,ii);
 			if(*have_mode) err_dma = 1;
@@ -1134,7 +1152,7 @@ void print_ascii(__u16 *p, __u8 length) {
 	char cl;
 	
 	/* find first non-space & print it */
-	for(ii = 0; ii< length; ii++) {
+	for (ii = 0; ii< length; ii++) {
 		if(((char) 0x00ff&((*p)>>8)) != ' ') break;
 		if((cl = (char) 0x00ff&(*p)) != ' ') {
 			if(cl != '\0') printf("%c",cl);
@@ -1144,8 +1162,8 @@ void print_ascii(__u16 *p, __u8 length) {
 		p++;
 	}
 	/* print the rest */
-	for(; ii< length; ii++) {
-		unsigned char c;
+	for (; ii < length; ii++) {
+		__u8 c;
 		/* some older devices have NULLs */
 		c = (*p) >> 8;
 		if (c) putchar(c);
